package eduni.inf.nbmodules.pepa.editor;

import eduni.inf.nbmodules.pepa.cookies.ReflectionCookie;
import eduni.inf.nbmodules.pepa.cookies.StateSpaceCookie;
import eduni.inf.nbmodules.pepa.cookies.SyntaxCheckedCookie;
import eduni.inf.nbmodules.pepa.editor.parser.ParseException;
import eduni.inf.nbmodules.pepa.editor.parser.PepaParser;
import eduni.inf.nbmodules.pepa.loader.PepaDataObject;
import eduni.inf.pepa.pepa.Peparoni;
import eduni.inf.pepa.pepamodel.PepaModel;


import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import javax.swing.Timer;
import org.openide.ErrorManager;
import org.openide.cookies.*;
import org.openide.filesystems.FileLock;
import org.openide.filesystems.FileObject;



import org.openide.text.Annotatable;
import org.openide.text.Annotation;
import org.openide.text.DataEditorSupport;
import org.openide.text.Line;
import org.openide.text.DataEditorSupport.Env;
import org.openide.text.Line.Set;
import org.openide.windows.CloneableOpenSupport;







/** Provides editor support for PEPA.
 *
 * @author N.V.Haenel
 */

public class PepaEditorSupport extends DataEditorSupport implements EditorCookie, OpenCookie, CloseCookie, PrintCookie {
    
    /** Timer which countdowns the auto-reparsing time. */
    Timer timer;
    
    private static boolean BACKGROUND_PARSING = true;
    
    private transient boolean hasAnnotations = false;
    
    private PepaModel pepaModel;
    
    /* List of error annotations attached to this document */
    private ArrayList errorAnnotations=new ArrayList();
    
    /** Create a new editor support.
     * @param obj the data object whose primary file will be edited as text
     */
    public PepaEditorSupport(PepaDataObject obj) {
        super(obj, new PepaEnv(obj));
        // Set a MIME type as needed, e.g.:
        setMIMEType("text/x-pepa");
    }
    
    
    /** Called when the document is modified.
     * Here, adding a save cookie to the object and marking it modified.
     * @return true if the modification is acceptable
     */
    protected boolean notifyModified() {
        if (!super.notifyModified()) {
            return false;
        }
        PepaDataObject obj = (PepaDataObject)getDataObject();
        if (obj.getCookie(SaveCookie.class) == null) {
            obj.setModified(true);
            // You must implement this method on the object:
            obj.addSaveCookie(new Save());
        }
        // start background parsing
        if(BACKGROUND_PARSING)  this.restartTimer();
        this.syntaxUnchecked();
        return true;
    }
    
    public void open(){
        if(BACKGROUND_PARSING)  this.restartTimer();
        super.open();
    }
    
    public boolean close(){
        //this.clearAnnotations();
        return super.close();
    }
    
    public PepaModel getPepaModel(){
        PepaParser p = this.parseModel();
        if (p.getParseExceptions().isEmpty())
            return this.pepaModel;
        else{
            try{
                ByteArrayOutputStream baos = new ByteArrayOutputStream();
                PrintStream ps = new PrintStream(baos);
                ps.println("The PEPA parser encountered syntax errors");
                for(Iterator i = p.getParseExceptions().iterator();i.hasNext();){
                    ps.println(i.next());
                }
                ps.flush();
                ErrorManager.getDefault().notify(ErrorManager.EXCEPTION, new Exception(baos.toString()));
                ps.close();
            }catch(Exception ioe){
                ErrorManager.getDefault().notify(ErrorManager.EXCEPTION, ioe);                                                
            }            
        }
        return null;
        
        
    }
    
    /** Called when the document becomes unmodified.
     * Here, removing the save cookie from the object and marking it unmodified.
     */
    protected void notifyUnmodified() {
        PepaDataObject obj = (PepaDataObject)getDataObject();
        SaveCookie save = (SaveCookie)obj.getCookie(SaveCookie.class);
        if (save != null) {
            // You must implement this method on the object:
            obj.removeSaveCookie(save);
            obj.setModified(false);
        }
        super.notifyUnmodified();
    }
    
    /** Should be called when the document is modified
     *  in this case adding a syntax check cookie
     *  and removing other cookies.
     *
     */
    protected void syntaxUnchecked(){
        PepaDataObject obj = (PepaDataObject)getDataObject();
        if (obj.getCookie(SyntaxCheckedCookie.class) != null)
            obj.removeCookie(obj.getCookie(SyntaxCheckedCookie.class));
        if (obj.getCookie(StateSpaceCookie.class) != null)
            obj.removeCookie(obj.getCookie(StateSpaceCookie.class));
        if (obj.getCookie(ReflectionCookie.class) != null)
            obj.removeCookie(obj.getCookie(ReflectionCookie.class));
        
        Peparoni.initialise();
    }
    
    /** Should be called when the document is checked for syntax errors
     *  and none appear
     *  Removing the syntax check cookie and marking the document as checked
     */
    public void syntaxChecked(){
        PepaDataObject obj = (PepaDataObject)getDataObject();
        obj.addCookie(new SyntaxCheckedCookie());
        try{
            this.saveDocument();
        }catch(IOException ioe){
            
        }
    }
    
    /** A save cookie to use for the editor support.
     * When saved, saves the document to disk and marks the object unmodified.
     */
    private class Save implements SaveCookie {
        public void save() throws IOException {
            saveDocument();
            getDataObject().setModified(false);
        }
    }
    
    
    // If the timer is running, it is restarted
    // otherwise a new one is started
    void restartTimer() {
        // will need to get the delay from the nb-pepa properties sheet
        int delay = 3000;
        
        if (timer==null) {  // initialize timer
            timer = new Timer(0, new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    changeTimeoutElapsed();
                }
            });
            timer.setRepeats(false);
            timer.setInitialDelay(delay);
        }
        timer.restart();
    }
    
    private PepaParser parseModel(){
        InputStream input = null;
        try{
            input = this.getInputStream();
        }catch(IOException ioe){
            ErrorManager.getDefault().notify(ErrorManager.EXCEPTION, ioe);

        }
        PepaParser p = new PepaParser(input);
        p.parse();
        pepaModel = p.getModel();
        return p;
    }
    
    private void  changeTimeoutElapsed(){
        PepaParser p = this.parseModel();
        this.clearAnnotations();
        if(!p.getParseExceptions().isEmpty())
            updateAnnotations(p.getParseExceptions());
    }
    
    private void  updateAnnotations(List parseExceptions){
        PepaDataObject obj = (PepaDataObject)this.getDataObject();
        LineCookie cookie = (LineCookie)obj.getCookie(LineCookie.class);
        Set lineSet = this.getLineSet();
        
        ParseException e;
        for(Iterator i = parseExceptions.iterator(); i.hasNext();){
            e = (ParseException)i.next();
            final Line line = lineSet.getCurrent(e.getLineNumber()-1);
            final PepaParseExceptionAnnotation ann = new PepaParseExceptionAnnotation(e.getShortDescription());
            ann.attach(line);
            ann.moveToFront();
            this.errorAnnotations.add(ann);
            line.addPropertyChangeListener(new PropertyChangeListener() {
                public void propertyChange(PropertyChangeEvent ev) {
                    String type = ev.getPropertyName();
                    if (type == null || type == Annotatable.PROP_TEXT) {
                        // User edited the line, assume error should be cleared.
                        // as the document is modified
                        // the annotation is re-attached if
                        // the error was not corrected
                        ann.detach();
                        line.removePropertyChangeListener(this);
                    }
                }
            });
        }
        
        
    }
    
    
    
    
    /** Removes all annotations on this EditorSupport */
    private void clearAnnotations(){
        for(Iterator i = errorAnnotations.iterator();i.hasNext();){
            ((Annotation)i.next()).detach();
        }
        errorAnnotations = new ArrayList();
    }
    
    /** A description of the binding between the editor support and the object.
     * Note this may be serialized as part of the window system and so
     * should be static, and use the transient modifier where needed.
     */
    private static class PepaEnv extends Env {
        
        private static final long serialVersionUID = 1L;
        
        /** Create a new environment based on the data object.
         * @param obj the data object to edit
         */
        public PepaEnv(PepaDataObject obj) {
            super(obj);
        }
        
        /** Get the file to edit.
         * @return the primary file normally
         */
        protected FileObject getFile() {
            return getDataObject().getPrimaryFile();
        }
        
        /** Lock the file to edit.
         * Should be taken from the file entry if possible, helpful during
         * e.g. deletion of the file.
         * @return a lock on the primary file normally
         * @throws IOException if the lock could not be taken
         */
        protected FileLock takeLock() throws IOException {
            return ((PepaDataObject)getDataObject()).getPrimaryEntry().takeLock();
        }
        
        /** Find the editor support this environment represents.
         * Note that we have to look it up, as keeping a direct
         * reference would not permit this environment to be serialized.
         * @return the editor support
         */
        public CloneableOpenSupport findCloneableOpenSupport() {
            return (PepaEditorSupport)getDataObject().getCookie(PepaEditorSupport.class);
        }
        
    }
    
}
