package eduni.inf.nbmodules.pepa.editor;

import eduni.inf.nbmodules.pepa.loader.PepaDataObject;
import eduni.inf.nbmodules.pepa.loader.PepaSecondaryNode;
import java.io.IOException;
import org.openide.cookies.*;
import org.openide.filesystems.FileLock;
import org.openide.filesystems.FileObject;
import org.openide.text.DataEditorSupport;
import org.openide.text.DataEditorSupport.Env;
import org.openide.windows.CloneableOpenSupport;


/** Support for editing a data object's secondary entry.
 *
 *  @see PepaNetSecondrayNode
 *
 * @author N.V.Haenel
 */
public class PepaSecondaryEditor extends DataEditorSupport implements EditorCookie, OpenCookie, CloseCookie, PrintCookie {
    
    private PepaSecondaryNode psn;
    
    /** Create a new editor support.
     *
     * @param obj the Main file 
     * @param pnsn the node for which to create the editor support
     * @param mimeType MIME Type of the data in the node
     * 
     */
    public PepaSecondaryEditor(PepaDataObject obj, PepaSecondaryNode psn, String mimeType) {
        super(obj, new SecondaryEnv(obj, psn));
        // Set a MIME type as needed, e.g.:
        this.psn = psn;
        setMIMEType(mimeType);
        
    }
    
    /** Called when the document is modified.
     * Here, adding a save cookie to the object and marking it modified.
     * @return true if the modification is acceptable
     */
    protected boolean notifyModified() {
        if (!super.notifyModified()) {
            return false;
        }
        PepaDataObject obj = (PepaDataObject)getDataObject();
        if (obj.getCookie(SaveCookie.class) == null) {
            obj.setModified(true);
            // You must implement this method on the object:
            obj.addSaveCookie(new Save());
        }
        return true;
    }
    
    /** Called when the document becomes unmodified.
     * Here, removing the save cookie from the object and marking it unmodified.
     */
    protected void notifyUnmodified() {
        PepaDataObject obj = (PepaDataObject)getDataObject();
        SaveCookie save = (SaveCookie)obj.getCookie(SaveCookie.class);
        if (save != null) {
            // You must implement this method on the object:
            obj.removeSaveCookie(save);
            obj.setModified(false);
        }
        super.notifyUnmodified();
    }
    
    /** Constructs message that should be used to name the editor component.
     *
     * @return name of the editor
     */
    protected String messageName() {
        if (! super.getDataObject().isValid()) return ""; // NOI18N        
        return this.psn.getEntry().getFile().getNameExt();        
    }
    
    
          
    /** A save cookie to use for the editor support.
     * When saved, saves the document to disk and marks the object unmodified.
     */
    private class Save implements SaveCookie {
        
        public void save() throws IOException {
            saveDocument();
            getDataObject().setModified(false);
        }
    }
    
    /** A description of the binding between the editor support and the object.
     * Note this may be serialized as part of the window system and so
     * should be static, and use the transient modifier where needed.
     */
    private static class SecondaryEnv extends Env {
        
        private static final long serialVersionUID = 1L;
        
        private PepaSecondaryNode psn;
        
        /** Create a new environment based on the data object.
         * @param obj the data object to edit
         */
        public SecondaryEnv(PepaDataObject obj, PepaSecondaryNode psn) {
            super(obj);
            this.psn = psn;                        
        }
        
        /** Get the file to edit.
         */
        protected FileObject getFile() {
            //return getDataObject().getPrimaryFile();
            return psn.getEntry().getFile();            
        }
        
        
        /** Lock the file to edit.
         * Should be taken from the file entry if possible, helpful during
         * e.g. deletion of the file.
         * @return a lock on the primary file normally
         * @throws IOException if the lock could not be taken
         */
        protected FileLock takeLock() throws IOException {
            //return ((MyDataObject)getDataObject()).getPrimaryEntry().takeLock();
            return psn.getEntry().takeLock();            
        }
        
        /** Find the editor support this environment represents.
         * Note that we have to look it up, as keeping a direct
         * reference would not permit this environment to be serialized.
         * @return the editor support
         */
        public CloneableOpenSupport findCloneableOpenSupport() {
            return (PepaSecondaryEditor)psn.getCookie(PepaSecondaryEditor.class);
        }        
        
        
        
    }
    
}
