package eduni.inf.nbmodules.xmi.editor;

import java.awt.Image;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import org.openide.cookies.*;
import org.openide.filesystems.FileLock;
import org.openide.filesystems.FileObject;
import org.openide.loaders.DataNode;
import org.openide.loaders.DataObject;
import org.openide.text.CloneableEditor;
import org.openide.text.DataEditorSupport;
import org.openide.text.DataEditorSupport.Env;
import org.openide.windows.CloneableOpenSupport;

import eduni.inf.nbmodules.xmi.loaders.*;


/** Support for editing XMI files and proj files inside a ZUML archive.
 *
 * @author N.V.Haenel
 */
// Replace OpenCookie with EditCookie or maybe ViewCookie as desired:
public class ZumlEditorSupport extends DataEditorSupport implements EditorCookie, OpenCookie, CloseCookie, PrintCookie {
    
    public static final int XMI_MODE = 1;
    public static final int PROJ_MODE = 2;
    
    int mode;
    
    private Image icon;
    
    /** Create a new editor support.
     * @param obj the data object whose primary file will be edited as text
     */
    public ZumlEditorSupport(ZumlDataObject obj, int mode, Image icon) {
        super((DataObject)obj,  (Env)new ZumlEnv(obj, mode));
        // Set a MIME type as needed, e.g.:
        setMIMEType("text/xml");
        this.mode = mode;
        this.icon = icon;
    }
    
    /** Called when the document is modified.
     * Here, adding a save cookie to the object and marking it modified.
     * @return true if the modification is acceptable
     */
    public boolean notifyModified() {
        if (!super.notifyModified()) {
            return false;
        }
        ZumlDataObject obj = (ZumlDataObject)getDataObject();
               
        if(mode == XMI_MODE)
            obj.setXmiModified(true);
        else
            obj.setProjModified(true);
        
        if ((obj.getCookie(SaveCookie.class)) == null) {
            obj.setModified(true);
            // You must implement this method on the object:
            obj.addCookie(new Save());                        
        }
        return true;
    }
    
    /** Called when the document becomes unmodified.
     * Here, removing the save cookie from the object and marking it unmodified.
     */
    public void notifyUnmodified() {
        ZumlDataObject obj = (ZumlDataObject)getDataObject();
        SaveCookie save = (SaveCookie)obj.getCookie(SaveCookie.class);
        if (save != null) {
            if(mode == XMI_MODE)
                obj.setXmiModified(false);
            else
                obj.setProjModified(false);
        }
        super.notifyUnmodified();
    }
    
    
    protected void initializeCloneableEditor(CloneableEditor editor) {
        super.initializeCloneableEditor(editor);
        editor.setIcon(icon);
    }
    
    protected String messageName () {
        String name;
        if(DataNode.getShowFileExtensions()) {            
            name = this.getDataObject().getPrimaryFile().getName()+(mode == XMI_MODE?".xmi":".proj");
        } else {
            name = this.getDataObject().getPrimaryFile().getName();
        }
        return name;
    }
        
    
    public void saveDocument() throws IOException{
        super.saveDocument();
        ((ZumlDataObject)this.getDataObject()).serializeArchive();
    }
    
    /** A save cookie to use for the editor support.
     * When saved, saves the document to disk and marks the object unmodified.
     */
    private class Save implements SaveCookie {
        public void save() throws IOException {
            saveDocument();
        }
    }
    
    
    
    /** A description of the binding between the editor support and the object.
     * Note this may be serialized as part of the window system and so
     * should be static, and use the transient modifier where needed.
     */
    private static class ZumlEnv extends Env {
        
        private static final long serialVersionUID = 1L;
        
        private ZumlDataObject zdo;
        
        private int mode;
        
        
        /** Create a new environment based on the data object.
         * @param obj the data object to edit
         */
        public ZumlEnv(ZumlDataObject obj, int mode) {
            super(obj);
            zdo = obj;
            this.mode = mode;
        }
        
        /** Get the file to edit.
         * @return the primary file normally
         */
        protected FileObject getFile() {
            return getDataObject().getPrimaryFile();
        }
        
        public InputStream inputStream() throws IOException {
            // if we are editing XMI
            if(mode == XMI_MODE)
                return zdo.getXmiInputStream();
            else
                return zdo.getProjInputStream();
        }
        
        
        public OutputStream outputStream() throws IOException {
            if(mode == XMI_MODE)
                return zdo.getXmiOutputStream();
            else
                return zdo.getProjOutputStream();
        }
        
        
        /** Lock the file to edit.
         * Should be taken from the file entry if possible, helpful during
         * e.g. deletion of the file.
         * @return a lock on the primary file normally
         * @throws IOException if the lock could not be taken
         */
        protected FileLock takeLock() throws IOException {
            return ((ZumlDataObject)getDataObject()).getPrimaryEntry().takeLock();
        }
        
        /** Find the editor support this environment represents.
         * Note that we have to look it up, as keeping a direct
         * reference would not permit this environment to be serialized.
         * @return the editor support
         */
        public CloneableOpenSupport findCloneableOpenSupport() {
            return (ZumlEditorSupport)getDataObject().getCookie(ZumlEditorSupport.class);
        }
        
    }
    
}
