package eduni.inf.nbmodules.xmi.loaders;

import eduni.inf.nbmodules.xmi.editor.ZumlEditorSupport;
import eduni.inf.pepa.extractor.PEPA_Extractor2;
import java.io.*;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.zip.ZipEntry;
import java.util.zip.ZipException;
import java.util.zip.ZipFile;
import java.util.zip.ZipOutputStream;
import javax.swing.JEditorPane;
import org.openide.ErrorManager;
import org.openide.cookies.SaveCookie;
import org.openide.filesystems.FileObject;
import org.openide.filesystems.FileStateInvalidException;
import org.openide.filesystems.FileUtil;
import org.openide.loaders.DataObjectExistsException;
import org.openide.loaders.MultiDataObject;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import org.openide.nodes.Children.Array;
import org.openide.nodes.Node.Cookie;
import org.openide.util.HelpCtx;


/** Represents a ZUML archive in the Repository.
 *
 * @author N.V.Haenel
 */
public class ZumlDataObject extends MultiDataObject {
    // should have two fields, one for the xmi and one for the project
    
    private ZipFile zip;
    private ZipEntry xmi;
    private ZipEntry proj;
    
    private String xmiName;
    private String projName;
    
    private ByteArrayOutputStream xmiOutputBuffer;
    private ByteArrayOutputStream projOutputBuffer;
    
    private byte[] xmiBuffer;
    private byte[] projBuffer;
    
    private boolean xmiModified;
    private boolean projModified;
    
    private ZumlSecondaryNode zumlXmiNode;
    private ZumlSecondaryNode zumlProjNode;
    
    private boolean sync;
    
    
    public ZumlDataObject(FileObject pf, ZumlDataLoader loader) throws DataObjectExistsException {
        super(pf, loader);
        init();
    }
    
    private void init() {
        sync = false;
        this.resetZipStream();
        
    }
    public void resetZipStream(){
        try{
            // get the zip file
            File filef = FileUtil.toFile(this.getPrimaryFile());
            if(filef != null){
                zip = new ZipFile(filef);
                
                for(Enumeration e = zip.entries();e.hasMoreElements();){
                    ZipEntry next = (ZipEntry)e.nextElement();
                    if(next.getName().endsWith(".xmi")){
                        xmi = next;
                        xmiName = xmi.getName();
                    }else if(next.getName().endsWith(".proj")){
                        // must be the project entry store as that
                        proj = next;
                        projName = proj.getName();
                    }
                }
                
                ZipEntry[] entries = new ZipEntry[]{xmi,proj};
                Object[]   buffers = new Object[]{xmiBuffer, projBuffer};
                
                for(int i = 0;i<entries.length;i++){
                    if(entries[i] != null){
                        InputStream is = zip.getInputStream(entries[i]);
                        BufferedInputStream  bis = new BufferedInputStream(is);
                        ByteArrayOutputStream baos = new ByteArrayOutputStream();
                        BufferedOutputStream bos = new BufferedOutputStream(baos);
                        byte[] data = new byte[10240];
                        while (bis.available() != 0) {
                            int read = bis.read(data, 0, data.length);
                            if (read == -1)
                                break;
                            bos.write(data, 0, read);
                        }
                        buffers[i] = baos.toByteArray();
                    }
                }
                
                xmiBuffer = (byte[])buffers[0];
                projBuffer = (byte[])buffers[1];
                
                zip.close();
                
                Node[] nodes = this.getChildren().getNodes();
                for(int i = 0;i<nodes.length;i++){
                    ((ZumlEditorSupport)nodes[i].getCookie(ZumlEditorSupport.class)).close();
                }
                this.getPrimaryFile().getFileSystem().refresh(true);
                
                ((ZumlDataNode)this.getNodeDelegate()).updateChildren(this.getChildren());;                
            }
            
        }catch(FileStateInvalidException fsie){
            ErrorManager.getDefault().notify(ErrorManager.EXCEPTION, fsie);
        }catch(ZipException ze){
            ErrorManager.getDefault().notify(ErrorManager.EXCEPTION, ze);
        }catch(IOException ioe){
            ErrorManager.getDefault().notify(ErrorManager.EXCEPTION, ioe);
            
        }
    }
    
    public void setXmiModified(boolean modified){
        if(this.xmiModified != modified){
            this.xmiModified = modified;
            this.firePropertyChange(this.PROP_MODIFIED,
            !modified ? Boolean.TRUE : Boolean.FALSE,
            modified ? Boolean.TRUE : Boolean.FALSE);
        }
    }
    
    public void setProjModified(boolean modified){
        if(this.projModified != modified){
            this.projModified = modified;
            this.firePropertyChange(this.PROP_MODIFIED,
            !modified ? Boolean.TRUE : Boolean.FALSE,
            modified ? Boolean.TRUE : Boolean.FALSE);
        }
    }
    
    public boolean isModified(){
        return (super.isModified() || this.xmiModified || this.projModified);
    }
    
    public InputStream getXmiInputStream() throws IOException{
        return new ByteArrayInputStream(xmiBuffer);
    }
    
    public InputStream getProjInputStream() throws IOException{
        return new ByteArrayInputStream(projBuffer);
    }
    
    public OutputStream getXmiOutputStream() throws IOException{
        xmiOutputBuffer = new ByteArrayOutputStream();
        return xmiOutputBuffer;
    }
    
    public OutputStream getProjOutputStream() throws IOException{
        projOutputBuffer = new ByteArrayOutputStream();
        return projOutputBuffer;
    }
    
    /** Invoke when the contents of the archive has changed
     */
    public synchronized void serializeArchive() throws IOException{
        
        if(!sync){
            sync = true;
            
            if(this.projModified)
                ((ZumlEditorSupport)zumlProjNode.getCookie(ZumlEditorSupport.class)).saveDocument();
            if(this.xmiModified)
                ((ZumlEditorSupport)zumlXmiNode.getCookie(ZumlEditorSupport.class)).saveDocument();
            
            this.setModified(false);
            ((ZumlEditorSupport)zumlProjNode.getCookie(ZumlEditorSupport.class)).notifyUnmodified();
            ((ZumlEditorSupport)zumlXmiNode.getCookie(ZumlEditorSupport.class)).notifyUnmodified();
            
            
            
            SaveCookie save = (SaveCookie)this.getCookie(SaveCookie.class);
            if(save != null)
                this.removeCookie(save);
            
            // this is a buffer for the new archive
            // write raw bytes into this buffer
            // as we cannot overwrite the old file yet, as we may have to read from
            // it still.
            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            
            ZipOutputStream zos = new ZipOutputStream(new BufferedOutputStream(bos));
            
            zos.putNextEntry(new ZipEntry(xmi.getName()));
            if(xmiOutputBuffer != null)
                xmiOutputBuffer.writeTo(zos);
            else{
                BufferedInputStream bis = new BufferedInputStream(new ByteArrayInputStream(xmiBuffer));
                
                byte[] data = new byte[10240];
                while (bis.available() != 0) {
                    int read = bis.read(data, 0, data.length);
                    if (read == -1)
                        break;
                    zos.write(data, 0, read);
                }
                bis.close();
            }
            zos.closeEntry();
            
            
            zos.putNextEntry(new ZipEntry(proj.getName()));
            if(projOutputBuffer != null)
                projOutputBuffer.writeTo(zos);
            else{
                BufferedInputStream bis = new BufferedInputStream(new ByteArrayInputStream(projBuffer));
                
                byte[] data = new byte[10240];
                
                while (bis.available() != 0) {
                    int read = bis.read(data, 0, data.length);
                    if (read == -1)
                        break;
                    zos.write(data, 0, read);
                }
                bis.close();
            }
            zos.closeEntry();
            
            zos.flush();
            zos.close();
            
            sync = false;
            
            // now we can write the new buffer to file
            // this creates a new file
            // maybe this can be done by updating the archive instead
            bos.writeTo(new FileOutputStream(zip.getName()));
            
            this.resetZipStream();
            this.xmiOutputBuffer = null;
            this.projOutputBuffer = null;
        }
    }
    
    
    public HelpCtx getHelpCtx() {
        return HelpCtx.DEFAULT_HELP;
        // If you add context help, change to:
        // return new HelpCtx(XmiDataObject.class);
    }
    
    protected Node createNodeDelegate() {
        return new ZumlDataNode(this, this.getChildren());
    }
    
    protected Children getChildren(){
        Children children = new Array();
        ArrayList nodes = new ArrayList();
        
        if(xmi != null){
            ZumlSecondaryNode zumlXmiNode = new ZumlSecondaryNode(this, "eduni/inf/nbmodules/xmi/resources/XmiDataIcon", ".xmi", ZumlEditorSupport.XMI_MODE, this.getUmlDiagrams());
            nodes.add(zumlXmiNode);
        }
        if(proj != null){
            ZumlSecondaryNode zumlProjNode = new ZumlSecondaryNode(this, "eduni/inf/nbmodules/xmi/resources/ProjDataIcon", ".proj", ZumlEditorSupport.PROJ_MODE);
            
            nodes.add(zumlProjNode);
        }
        
        Node[] nodeArray = new Node[nodes.size()];
        children.add((Node[])nodes.toArray(nodeArray));
        return children;
    }
    
    protected Children getUmlDiagrams(){
        Children children = new Array();
        ArrayList nodes = new ArrayList();
        
        PEPA_Extractor2 p = new PEPA_Extractor2();
        
        if(p.parse(FileUtil.toFile(this.getPrimaryFile()))){
            List c= p.getAllDiagramsStrings(Arrays.asList(new String[]{"CollaborationDiagram"}));
            for(Iterator i = c.iterator();i.hasNext();){
                nodes.add(new UmlDiagramNode(this, (String)i.next() , "CollaborationDiagram"));
            }
            List d= p.getAllDiagramsStrings(Arrays.asList(new String[]{"SequenceDiagram"}));
            for(Iterator i = d.iterator();i.hasNext();){
                nodes.add(new UmlDiagramNode(this, (String)i.next(), "SequenceDiagram"));
            }
            List e= p.getAllDiagramsStrings(Arrays.asList(new String[]{"ActivityDiagram"}));
            for(Iterator i = e.iterator();i.hasNext();){
                nodes.add(new UmlDiagramNode(this, (String)i.next() , "ActivityDiagram"));
            }
            
        }
        
        
        Node[] nodeArray = new Node[nodes.size()];
        children.add((Node[])nodes.toArray(nodeArray));
        return children;
        
    }
    
    
    public final void addCookie(Cookie cookie) {
        getCookieSet().add(cookie);
    }
    
    public final void removeCookie(Cookie cookie) {
        getCookieSet().remove(cookie);
    }
    
    
}
