package eduni.inf.pepa.pepamodel.other;

import java.util.*;

/** Utility class that represents sorted action sets.
 *
 *  <p>This class favours composition over inheritance, and therefore doesn't
 *     extend <code>TreeSet</code>, but instead keeps a member instance and forwards
 *     all method invocations to that member. This class is entirely parametric,
 *     which means that only <code>PepaAction</code> objects may be added. The provided
 *     <code>Iterator</code> therefore guarantees that all returned objects may safely
 *     be cast to <code>PepaAction</code> objects at runtime. Any methods that remove or 
 *     delete elements are optional and this implementation throws 
 *     <code>UnsupportedOperationException</code>s. Also note that this class
 *     does not support <code>Comparator</code>s, because <code>PepaAction</code>
 *     implements <code>Comparable</code>.
 *  </p>
 *      
 *
 *
 * @author  N.V.Haenel
 * @since 0.9.5
 */
public class ActionSet implements Comparable, SortedSet{
    private TreeSet actionSet;
    
    public ActionSet() {
        this.actionSet = new TreeSet();
    }
    public ActionSet(Comparator c) {
        throw new UnsupportedOperationException("ActionSet does not support Comparators");
    }
    public ActionSet(SortedSet s){
        // need to check they are actions        
        this.actionSet = new TreeSet(s);
    }
    public ActionSet(Collection c){
        this();
        actionSet.addAll(c);
    }
    public ActionSet(PepaAction[] actions) { 
        this();
        actionSet.addAll(Arrays.asList(actions));        
    }
    
    public int compareTo(Object obj) {
        ActionSet that = (ActionSet)obj;
        int k;
        for(Iterator i = this.iterator(), j = that.iterator();;){
            // loop breakers            
            if (!i.hasNext() && !j.hasNext())
                return 0; // if end of both sets, and still no difference
            else if(!i.hasNext())
                return -1;// if end of first set, but still no difference and second set is larger
            else if(!j.hasNext())
                return 1;// if end of second set, but still no difference and first set is larger  
            
            if((k = ((PepaAction)i.next()).compareTo((PepaAction)j.next()))!=0)                                
                return k;                      
        }
        
    }
    
    private ArrayList confirmTypes(Collection c){
        ArrayList actions = new ArrayList();
        for(Iterator i = c.iterator();i.hasNext();){
            actions.add((PepaAction)i.next());
        }
        return actions;
    }
    
    public boolean add(Object obj) {
        return actionSet.add((PepaAction)obj);
    }
    
    public boolean addAll(Collection collection) {
        return actionSet.addAll(confirmTypes(collection));
    }
    
    public void clear() {
        throw new UnsupportedOperationException("ActionSet does not support clear()");
    }
    
    public Comparator comparator() {
        return actionSet.comparator();
    }
    
    public boolean contains(Object obj) {
        return actionSet.contains((PepaAction)obj);
        
    }
    
    public boolean containsAll(Collection collection) {
        return actionSet.containsAll(confirmTypes(collection));
    }
    
    public Object first() {
        return actionSet.first();
    }
    
    public SortedSet headSet(Object obj) {
        return actionSet.headSet((PepaAction)obj);
    }
    
    public boolean isEmpty() {
        return actionSet.isEmpty();
    }
    
    public Iterator iterator() {
        return actionSet.iterator();
    }
    
    public Object last() {
        return actionSet.last();
    }
    
    public boolean remove(Object obj) {
        return actionSet.remove(obj);
    }
    
    public boolean removeAll(Collection collection) {
        return actionSet.removeAll(collection);
    }
    
    public boolean retainAll(Collection collection) {
        return actionSet.retainAll(collection);
    }
    
    public int size() {
        return actionSet.size();
    }
    
    public SortedSet subSet(Object obj1, Object obj2) {
        return actionSet.subSet((PepaAction)obj1, (PepaAction)obj2);
    }
    
    public SortedSet tailSet(Object obj) {
        return actionSet.tailSet((PepaAction)obj);
    }
        
    public Object[] toArray() {
        return actionSet.toArray();
    }
    
    public Object[] toArray(Object[] obj) {
        return actionSet.toArray(obj);
    }
    
    public String toString(){
        String s = actionSet.toString();
        s = s.replace('[','{');
        s = s.replace(']','}');
        return s;    
    }
    
    public static ActionSet union(ActionSet a, ActionSet b){
        ActionSet returnSet = new ActionSet(a);
        returnSet.addAll(b);
        return returnSet;        
    }
    
    public static ActionSet intersection(ActionSet a, ActionSet b){
        ActionSet returnSet = new ActionSet(a);
        returnSet.retainAll(b);
        return returnSet;                    
    }
    
    public static ActionSet difference(ActionSet a, ActionSet b){
        ActionSet returnSet = new ActionSet(a);
        returnSet.removeAll(b);
        return returnSet;                                    
    }
    
    
}
