package eduni.inf.pepa.pepamodel.other;

import eduni.inf.pepa.pepamodel.process.PepaVariable;
import eduni.inf.pepa.pepamodel.rate.PepaRate;
import java.util.ArrayList;
import java.util.List;


/** Represents PEPA Derivatives.
 *
 *  <p> Derivatives hold the following information:<br>
 *  <ol><li>The currently enabled action types. Those actions currently enabled by this process
 *      <li>The rate variables associated with these actions.
 *      <li>The identifier, as <code>PepaVariable</code>, of the target process for the actions.
 *      <li>The identifier, as <code>PepaVariable</code> of this derivative.
 *  </ol>
 *  </p>
 *
 * <p> The three arrays <code>actionTypes</code>, <code>rates</code> and <code>targets</code>
 *     should have the same length, and can be initialised through the constructor only. No accessor methods
 *     are provided for setting the values.
 * </p>
 *
 * <p> There is a unique natural ordering of derivatives. This holds only under the assumption
 *     that no PEPA model will have derivatives with the same name precisely. The 
 *     <code>compareTo()</code> compares <b>only</b> the <code>derivative</code> member
 *     of class <code>PepaVariable</code>. 
 * </p>
 *
 * <p><b> Be aware that this class has a natural ordering that is inconsistent with equals. </b></p>
 *
 *
 * @author  N.V.Haenel
 * @since 0.9.5
 */
public class PepaDerivative implements Comparable{
    
    /** The enabled one step action types for this derivative */
    private PepaAction[] actionTypes;
    /** the rates for each of the enabled action types */
    private PepaRate[] rates;
    /** The target single step derivatives as <code>PepaVariable</code>.*/
    private PepaVariable[] targets;
    /** The <code>PepaVariable</code> that labels this derivative */
    private PepaVariable derivative;
    
    /** Creates a new instance of <code>PepaDerivative</code> */
    public PepaDerivative( PepaVariable derivative, PepaAction[] actionTypes, PepaRate[] rates,  PepaVariable[] targets) {
        this.actionTypes = actionTypes;
        this.rates = rates;
        this.targets = targets;
        this.derivative = derivative;      
    }
    /** Standard accessor.
     *
     *  @return the enabled action types for this derivative     
     */
    public PepaAction[] getActionTypes(){
        return actionTypes;
    }
    
    /** Standard accessor.
     *
     *  @return the rates associated with the action types
     */
    public PepaRate[] getRates(){
        return rates;
    }
    
    /** Standard accessor.
     *
     *  @return the targets reached when executing the action types
     */
    public PepaVariable[] getTargets(){
        return targets;
    }
    
    public List getCurrentTargets(PepaAction a){
        ArrayList al = new ArrayList();
        for(int i = 0;i<actionTypes.length;i++){
            if(actionTypes[i].equals(a))
                al.add(targets[i]);
        }
        return al;
    }
    
    /** Standard accessor
     *
     * @return the identifier for this derivative
     */
    public PepaVariable getDerivative(){
        return derivative;
    }
    
    public boolean equals(Object o){
        return o instanceof PepaDerivative
        && ((PepaDerivative)o).getActionTypes().equals(this.getActionTypes())
        && ((PepaDerivative)o).getRates().equals(this.getRates())
        && ((PepaDerivative)o).getTargets().equals(this.getTargets())
        && ((PepaDerivative)o).getDerivative().equals(this.getDerivative());
    }
    
    public int hashCode(){
        int result = 17;
        result = 37*result + actionTypes.hashCode();
        result = 37*result + rates.hashCode();
        result = 37*result + targets.hashCode();
        result = 37*result + derivative.hashCode();
        return result;
    }
    
    public String toString(){
        StringBuffer b = new StringBuffer();
        b.append(derivative.getIdentifier());
        b.append(": ");
        for(int i = 0 ; i < actionTypes.length;i++){
            b.append("[");
            b.append(actionTypes[i].toString());
            b.append(",");
            b.append(rates[i].toString());
            b.append(",");
            b.append(targets[i].toString());
            b.append("] ");
        }
        return b.toString();
    }
    
    
    public int compareTo(Object o){       
        // ClassCastExcpetion throwen during cast
        if(this.equals(o))
            return 0;
        else 
            return this.getDerivative().compareTo(((PepaDerivative)o).getDerivative());
        
    }
 
}
