package eduni.inf.pepa.pepamodel.process;

import eduni.inf.pepa.pepamodel.visitors.PepaProcessVisitor;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/** Represents PEPA Choice objects.
 *
 * <p>This object has been implemented as a list of
 *    <code>PepaProcess</code> objects. This is
 *    contrary to the definition in Janes thesis
 *    where Choice is a binary operator. This makes it
 *    much easier to find the single step derivatives,
 *    and seems to be a more natural method of
 *    representation.
 * </p>
 *
 * <p> The class always maintains a natural ordering
 *     of its components, therefore any <code>PepaProcess</code>
 *     Objects should implement </code>Comparable</code>.
 * </p>
 *
 * @see java.lang.Comparable
 *
 * @author N.V.Haenel
 * @since 0.9.5
 *
 */
public class PepaChoice extends PepaProcess {
    /** Sigma of </code>PepaProcess<code> objects. */
    private List choices;
    
    /** Standrad Constructor */
    public PepaChoice(List choices){
        this.choices = new ArrayList(choices);
        Collections.sort(this.choices);
        
    }
    
    public void accept(PepaProcessVisitor visitor){
        visitor.visit(this);
    }
    
    /** Standard Accessor */
    public List getChoices(){
        return new ArrayList(choices);
    }
    
    public boolean equals(Object o){
        if (o instanceof PepaChoice){
            for(Iterator i = choices.iterator(), j =((PepaChoice)o).getChoices().iterator();;){
                // no need for casting as equals is inherited from Object
                if (!i.hasNext() && !j.hasNext())
                    return true; // if end of both sets, and still no difference
                else if(!i.hasNext())
                    return false;// if end of first set, but still no difference and second set is larger
                else if(!j.hasNext())
                    return false;// if end of second set, but still no difference and first set is larger
                
                if(!(i.next()).equals(j.next()))
                    return false;
            }
            
        }else return false;
    }
    
    public int hashCode(){
        int result = 17;
        for(Iterator i = choices.iterator();i.hasNext();){
            // no need for casting as hashCode is inherited from Object
            result = 37*result + i.next().hashCode();
        }
        return result;
    }
    
    public String toString(){
        StringBuffer b = new StringBuffer();
        b.append("(");
        for(Iterator i = choices.iterator();i.hasNext();){
            b.append(i.next().toString());
            if(i.hasNext()) b.append(" + ");
        }
        b.append(")");
        return b.toString();
    }
    
    public int compareTo(Object o){
        if(o instanceof PepaChoice){
            PepaChoice that = (PepaChoice)o;
            int k;
            for(Iterator i = this.getChoices().iterator(), j = that.getChoices().iterator();;){
                if (!i.hasNext() && !j.hasNext())
                    return 0; // if end of both sets, and still no difference
                else if(!i.hasNext())
                    return -1;// if end of first set, but still no difference and second set is larger
                else if(!j.hasNext())
                    return 1;// if end of second set, but still no difference and first set is larger                
                
                if((k = ((Comparable)i.next()).compareTo((Comparable)j.next()))!=0)
                    return k;
            }
            
        }
        else
            return super.compareTo(o);
    }
}

