package eduni.inf.pepa.pepamodel.process;

import eduni.inf.pepa.pepamodel.visitors.PepaProcessVisitor;


/** Represents a PEPA Process object.
 *
 *  <p>This abstract class is the inheritence base for the 5 types
 *     of PEPA component. Importantly this class provides a <code>compareTo()</code>
 *     method that is used to compare the relative ordering or Components.
 *     For example when comparing a Choice with a Cooperation.
 *     The Process Objects themselves each implement a <code>compareTo()</code>
 *     that is used to compare Processes of the same type, if they are
 *     different <code>super.compareTo()</code> is invoked
 *  </p>
 *
 * <p>The relative order is:
 * <code>
 * <ol><li>PepaVariable
 *     <li>PepaChoice
 *     <li>PepaPrefix
 *     <li>PepaCooperation
 *
 *     <li>PepaHide
 * </ol>
 * </code>
 *
 * @see PepaVariable
 * @see PepaChoice
 * @see PepaPrefix
 * @see PepaHide
 * @see PepaCooperation
 *
 * @author N.V.Haenel
 * @since 0.9.5
 *
 */

public abstract class PepaProcess implements PepaProcessVisitable, Comparable{
    public abstract void accept(PepaProcessVisitor visitor);
    
    /** Provide relative comparison between Processes.
     * <p>The relative order is:
     * <code>
     * <ol><li>PepaVariable
     *     <li>PepaChoice
     *     <li>PepaPrefix
     *     <li>PepaCooperation
     *     <li>PepaHide
     *
     * </ol>
     * </code>
     *
     */
    public int compareTo(Object o){
        // note anything returning a zero will never be called
        if(o instanceof PepaVariable){
            if(this instanceof PepaVariable){
                return 0;
            }
            else if(this instanceof PepaPrefix){
                return 1;
            }
            else if(this instanceof PepaChoice){
                return 1;
            }
            else if(this instanceof PepaCooperation){
                return 1;
            }
            else if(this instanceof PepaHide){
                return 1;
            }
        }
        else if(o instanceof PepaPrefix){
            if(this instanceof PepaVariable){
                return -1;
            }
            else if(this instanceof PepaPrefix){
                return 0;
            }
            else if(this instanceof PepaChoice){
                return 1;
            }
            else if(this instanceof PepaCooperation){
                return 1;
            }
            else if(this instanceof PepaHide){
                return 1;
            }
        }
        else if(o instanceof PepaChoice){
            if(this instanceof PepaVariable){
                return -1;
            }
            else if(this instanceof PepaPrefix){
                return -1;
            }
            else if(this instanceof PepaChoice){
                return 0;
            }
            else if(this instanceof PepaCooperation){
                return 1;
            }
            else if(this instanceof PepaHide){
                return 1;
            }
        }
        else if(o instanceof PepaCooperation){
            if(this instanceof PepaVariable){
                return -1;
            }
            else if(this instanceof PepaPrefix){
                return -1;
            }
            else if(this instanceof PepaChoice){
                return -1;
            }
            else if(this instanceof PepaCooperation){
                return 0;
            }
            else if(this instanceof PepaHide){
                return 1;
            }
        }
        else if(o instanceof PepaHide){
            if(this instanceof PepaVariable){
                return -1;
            }
            else if(this instanceof PepaPrefix){
                return -1;
            }
            else if(this instanceof PepaChoice){
                return -1;
            }
            else if(this instanceof PepaCooperation){
                return -1;
            }
            else if(this instanceof PepaHide){
                return 0;
            }
        }
        throw new ClassCastException(o.getClass().getName());
        
    }
}

