

package eduni.inf.pepa.pepamodel.statespace;

import eduni.inf.pepa.pepamodel.PepaModel;
import eduni.inf.pepa.pepamodel.other.DerivativeSet;
import eduni.inf.pepa.pepamodel.other.PepaAction;
import eduni.inf.pepa.pepamodel.other.PepaDerivative;
import eduni.inf.pepa.pepamodel.process.PepaCooperation;
import eduni.inf.pepa.pepamodel.process.PepaProcess;
import eduni.inf.pepa.pepamodel.process.PepaVariable;
import eduni.inf.pepa.pepamodel.visitors.DerivativeFinder;
import eduni.inf.pepa.pepamodel.visitors.ModelNormaliser;
import java.util.*;


/** This is the data structure for exploration of the state space of a PEPA model.
 *
 *  <p>The Explorer consists of a tree, where each leaf is a list of derivatives.
 *     Exploration of the state space beginns with the construction of the
 *     explorer, and the initialiosation of the leaf nodes. Then for each state, the
 *     current transitions are established in a nested bredth first search manner.
 *     Hence all the states that can be reached from the current state are identified,
 *     then all the duplicates are eliminated, and the found states are added to the
 *     unexplored list, and the current state is added to the explored list. Now the next state is
 *     fetched from the unexplored list, and designated the current state. Exploration continues
 *     until the unexplored list contains no items.
 *  </p>
 *
 * @author  N.V.Haenel
 */
public class StateSpaceExplorer {
    /** All the leaf nodes */
    private ArrayList explorerLeafs;
    /** The root of the Explorer */
    private ExplorerNode root;
    /** The PEPA model represented by the Explorer */
    private PepaModel model;
    /** The normed PEPA model represented by the Explorer */
    private PepaModel normed;
    /** The Derivative Sets for the PEPA model */
    private List derivativeSets;
    /** List of explored states */
    private LinkedList unexplored;
    /** List of unexplored states */
    private LinkedList explored;
    /** Running total of the number of states */
    private int stateCount;
    
    /** Creates a new instance of StateSpaceGenerator */
    public StateSpaceExplorer(PepaModel model) {
        // first norm the model
        this.model = model;
        ModelNormaliser mn = new ModelNormaliser();
        normed = mn.normaliseModel(model);
        // now find the derivative sets
        DerivativeFinder df = new DerivativeFinder();
        derivativeSets = df.findDerivatives(normed);
        // get the initial state
        PepaCooperation system = (PepaCooperation)normed.getSystemEquation();
        // initialise the leaf nodes
        explorerLeafs = new ArrayList();
        // generate the explorer
        root = (ExplorerNode)this.generateExplorer(system);
        // trim the size of the list of leaf pointers
        explorerLeafs.trimToSize();
    }
    
    /** Used to retrieve the <code>DerivativeSet</code> for each <code>ExplorerLeaf</code>s.
     *
     * @param v the variable encountered at the leaf of the tree
     * @return the derivative set for the variable
     */
    private DerivativeSet getDerivativeSet(PepaVariable v){
        for(Iterator i = derivativeSets.iterator();i.hasNext();){
            DerivativeSet current = (DerivativeSet)i.next();
            if(current.contains(v))
                return current;
        }
        throw new NoSuchElementException("Derivative "+v+" cannot be found in derivative set");
    }
    
    /** Assembles the explorer tree and returns the root node
     *
     * @param p the system equation of the model
     * @return the root node of the explorer tree
     */
    private ExplorerComponent generateExplorer(PepaCooperation p){
        ExplorerComponent left;
        ExplorerComponent right;
        PepaProcess p1 = p.getProcess1();
        PepaProcess p2 = p.getProcess2();
        
        if(p1 instanceof PepaVariable){
            DerivativeSet s = getDerivativeSet((PepaVariable)p1);
            left = new ExplorerLeaf(s, s.indexOf((PepaVariable)p1), explorerLeafs.size());
            explorerLeafs.add(left);
        }else
            left = generateExplorer((PepaCooperation)p.getProcess1());
        
        if(p2 instanceof PepaVariable){
            DerivativeSet s = getDerivativeSet((PepaVariable)p2);
            right = new ExplorerLeaf(s,s.indexOf((PepaVariable)p2),explorerLeafs.size());
            explorerLeafs.add(right);
        }else
            right = generateExplorer((PepaCooperation)p.getProcess2());
        
        return new ExplorerNode(left, right, p.getCoopSet());
    }
    
    /** Returns the current pointers of the explorer tree leaves.
     *  This finds the current configuration, that
     *  is a list of integer indices, one for each leaf node.
     *
     * @return a <code>LinkedList</code> of integers,
     *  describing the current configuration of the explorer tree
     */
    private LinkedList getCurrentConfiguration(){
        LinkedList config = new LinkedList();
        for(Iterator i = explorerLeafs.iterator();i.hasNext();){
            config.add(new Integer(((ExplorerLeaf)i.next()).getCurrentIndex()));
        }
        return config;
    }
    
    
    /** Sets the pointers of the leaves in the explorer tree.
     *  This is used when a state is fetched from the unexplored queue
     *  and the explorer needs to be set to that state
     *
     * @param l a <code>LinkedList</code> that
     * describes a configuration of the explorer tree.
     */
    private void setCurrentConfiguration(LinkedList l){
        for(Iterator i = l.iterator(), j = explorerLeafs.iterator();i.hasNext();){
            ((ExplorerLeaf)j.next()).setCurrent(((Integer)i.next()).intValue());
        }
    }
    
    /** Explores the state space of the model, as described in the overview.
     *
     */
    public void exploreStateSpace(){
        // init lists
        unexplored = new LinkedList();
        explored = new LinkedList();
        // add the first state to unexplored
        unexplored.add(this.getCurrentConfiguration());
        
        while(!unexplored.isEmpty()){
            // if we have seen this state before, disregard it.
            if(explored.contains(unexplored.getFirst())){
                unexplored.removeFirst();
            }
            else{
                // get the next state
                LinkedList current = (LinkedList)unexplored.removeFirst();
                // set explorer to this state
                this.setCurrentConfiguration(current);
                // enable actions
                this.root.setEnabledActions();
                // find all actions that can be made
                LinkedList l = new LinkedList(root.getEnabledActions());
                int number = 0;
                // for each action determine transitions
                for(PepaAction a;!l.isEmpty();){
                    a=(PepaAction)l.removeFirst();
                    // get current set of leafs that can participate
                    ArrayList al = this.getCurrentStateExecutions(root,a);
                    // make all the moves and eliminate duplicates
                    List newStates = this.executeActions(al, a);
                    unexplored.addAll(newStates);
                }
                explored.addLast(current);
                System.out.print("|");
            }
        }
        System.out.println("State Space exploration complete");
        System.out.println("The model has: "+explored.size()+" states.");
        System.out.println(explored);
    }
    
    /** Utility method to get the Current state as String */
    
    public String getCurrentStateString(ExplorerComponent c, PepaAction a){
        if(c instanceof ExplorerLeaf) return String.valueOf((((ExplorerLeaf)c).getIndex()));
        else return "("+getCurrentStateString(((ExplorerNode)c).getLeft(), a) +
        (((ExplorerNode)c).getCooperationSet().contains(a)?"&":"|") +
        getCurrentStateString(((ExplorerNode)c).getRight(), a)+")";
    }
    
    
    /**************************************************************************
     ******WARNING - TRICKY CODE NEXT - WARNING********************************
     **************************************************************************/
    
    /** Returns a <code>List</code> of List</code>s, that contain the indexes
     *  of <code>ExplorerLeafs</code> that participate together to move to a new state.
     *
     */
    public ArrayList getCurrentStateExecutions(ExplorerComponent c, PepaAction a){
        ArrayList returnList = new ArrayList();
        
        if(c instanceof ExplorerLeaf) {
            if(((ExplorerLeaf)c).getEnabledActions().contains(a)){
                ArrayList l = new ArrayList();
                l.add(new Integer(((ExplorerLeaf)c).getIndex()));
                returnList.add(l);
                return returnList;
            }
            return returnList;
        }
        else if (((ExplorerNode)c).getEnabledActions().contains(a)){
            ArrayList left = getCurrentStateExecutions(((ExplorerNode)c).getLeft(), a);
            ArrayList right = getCurrentStateExecutions(((ExplorerNode)c).getRight(), a);
            if(((ExplorerNode)c).getCooperationSet().contains(a))
                return merge(left,right);
            else
                return concat(left,right);
        }
        return returnList;
    }
    
    
    /** Takes a list of lists of executing leaves and retruns the new states.
     *
     * @param al a list of lists from <code>getCurrentStateExecutions()</code>
     * @param a the action to be executed
     *
     * @return a list of new states
     */
    private List executeActions(ArrayList al, PepaAction a){
        LinkedList newStates = new LinkedList();
        for(Iterator i = al.iterator();i.hasNext();){
            ArrayList l = (ArrayList)i.next();
            newStates.addAll(execute(l,a));
        }
        return newStates;
    }
    
    /** Takes a list of executing leaves, and returns new states.
     *
     * @param al a list of executing leaves
     * @param a the action being executed
     *
     * @return a list of new states
     */
    private List execute(ArrayList al, PepaAction a){
        ArrayList indexes = new ArrayList(explorerLeafs.size());
        for(int i = 0;i<explorerLeafs.size();i++ ){
            if(al.contains(new Integer(i))){
                indexes.add(getNewIndex((ExplorerLeaf)explorerLeafs.get(i),a));
            }else
                indexes.add(new Integer(((ExplorerLeaf)explorerLeafs.get(i)).getCurrentIndex()));
        }
        LinkedList list = getNewStates(new LinkedList(indexes));
        return list;
    }
    
    
    private List getNewIndex(ExplorerLeaf l, PepaAction a){
        DerivativeSet s = l.getDerivativeSet();
        PepaDerivative d = l.getCurrent();
        List targets = d.getCurrentTargets(a);
        ArrayList indexes = new ArrayList(targets.size());
        for(Iterator i = targets.iterator();i.hasNext();){
            indexes.add(new Integer(s.indexOf((PepaVariable)i.next())));
        }
        return indexes;
    }
    
    private LinkedList getNewStates(LinkedList l){
        if(l.size() == 0){
            LinkedList list = new LinkedList();
            list.add(new LinkedList());
            return list;
        }
        else if(l.getFirst() instanceof Integer){
            Integer i = (Integer)l.removeFirst();
            return addToAll(getNewStates(l),i);
        }else if (l.getFirst() instanceof ArrayList){
            ArrayList al = (ArrayList)l.removeFirst();
            LinkedList returnList = new LinkedList();
            LinkedList precopy = getNewStates(l);
            for(Iterator i = al.iterator();i.hasNext();){
                LinkedList copy = new LinkedList();
                for(Iterator j = precopy.iterator();j.hasNext();){
                    copy.add(((LinkedList)j.next()).clone());
                }
                returnList.addAll(addToAll(copy,(Integer)i.next()));
            }
            return returnList;
        }
        return null;
    }
    
    /** Utility method to add an integer to start of each list in a list of lists.
     *
     */
    private static LinkedList addToAll(LinkedList one, Integer in){
        
        for(Iterator i = one.iterator();i.hasNext();){
            LinkedList two = (LinkedList)i.next();
            two.addFirst(in);
        }
        return one;
    }
    
    
    /**************************************************************************
     ******WARNING - TRICKY CODE END - WARNING*********************************
     **************************************************************************/
    
    /** Utilits method to merge two <code>ArrayList</code>s. */
    public static ArrayList merge(ArrayList one, ArrayList two){
        ArrayList returnList = new ArrayList();
        
        if(one.isEmpty())  return two;
        if(two.isEmpty())  return one;
        
        ArrayList currentI;
        ArrayList currentJ;
        
        for(Iterator i = one.iterator();i.hasNext();){
            currentI = (ArrayList)i.next();
            for(Iterator j = two.iterator();j.hasNext();){
                currentJ = (ArrayList)j.next();
                currentI.addAll(currentJ);
                returnList.add(currentI);
            }
        }
        return returnList;
    }
    /** Utility method to concatinate two ArrayLists */
    public static ArrayList concat(ArrayList one, ArrayList two){
        ArrayList returnList = new ArrayList();
        returnList.addAll(one);
        returnList.addAll(two);
        return returnList;
    }
    
    
    public String toString(){
        return root.toString();
    }
}
