package eduni.inf.pepa.pepamodel.visitors;

import eduni.inf.pepa.pepamodel.PepaModel;
import eduni.inf.pepa.pepamodel.other.DerivativeSet;
import eduni.inf.pepa.pepamodel.other.PepaAction;
import eduni.inf.pepa.pepamodel.other.PepaDefinition;
import eduni.inf.pepa.pepamodel.other.PepaDerivative;
import eduni.inf.pepa.pepamodel.process.*;
import eduni.inf.pepa.pepamodel.rate.PepaRate;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;


/** Responsible for grouping together recursive definitions.
 *  This iterates through a given list of definitions
 *  and returns a <code>List</code> of <code>DerivativeSet</code>s.
 *  Each <code>DerivativeSet</code> contains <code>PepaDerivative</code>s
 *  that are recursively defined. Expects the model to be normalised
 *  and especially requires there to be no cooperation or, hides
 *  in the definitions.
 *
 * @author N.V.Haenel
 * @since 0.9.5
 */

public class DerivativeFinder {
    /** The model this <code>DerivativeFinder</code> is working on. */
    private PepaModel model;
    /** The definitions as a <code>HashMap</code> for efficient lookup of definitions */
    private HashMap definitions;
    /** The <code>List</code> of <code>DerivativeSet</code>s. */
    private List derivativeSets;
    
    
    /** Default constructor to fascilitate innstance reuse */
    public DerivativeFinder(){
    }
    
    /** Used to find the derivative sets of a <code>PepaModel</code>.
     *
     *  @param model the <code>PepaModel</code> for which the derivative sets should be found.
     *  @return a <code>List</code> of <code>TreeSet</code> objects.
     *
     */
    public List findDerivatives(PepaModel model){
        this.model = model;
        this.definitions = model.getDefinitionsHashMap();
        derivativeSets = new ArrayList();
        PepaDefinition current;
        DerivativesVisitor visitor = new DerivativesVisitor();
        for(Iterator i = model.getDefinitions().iterator();i.hasNext();){
            current = (PepaDefinition)i.next();
            if(!this.contains(current.getVariable())){
                derivativeSets.add(visitor.findDerivatives(current));
            }
        }
        return derivativeSets;
    }
    
    /** Used to check if the current definition is already contained within
     *  one of the derivatives sets.
     *
     */
    private boolean contains(PepaVariable var){
        DerivativeSet currentSet;
        PepaVariable current;
        for(Iterator i = derivativeSets.iterator();i.hasNext();){
            currentSet = (DerivativeSet)i.next();
            for(Iterator j = currentSet.iterator();j.hasNext();){
                current = (PepaVariable)((PepaDerivative)j.next()).getDerivative();
                if(current.equals(var))
                    return true;
            }
        }
        return false;
    }
    /** Used by the <code>DerivativeFinder</code> to find the derivatives set
     *  of a definition.
     */
    public class DerivativesVisitor extends AbstractPepaProcessVisitor{
        /** DerivativeSet for the initial definition */
        private DerivativeSet derivativeSet;
        /** Current definition identifier */
        private PepaVariable current;
        
        /** Default constructor used to fascilitate object reuse. Innstantiate
         *  the <code>DerivativesVisitor</code> once and then use on multiple
         *  <code>PepaDefinition</code>s.
         *
         *  Importantly we are assuming that the whole model has been normed.
         */
        public DerivativesVisitor(){
        }
        
        /** Used to find the derivatives of a single definition */
        DerivativeSet findDerivatives(PepaDefinition definition){
            // store initial definition identifier            
            this.current = definition.getVariable();
            // make new hash set
            this.derivativeSet = new DerivativeSet();
            definition.getDefinition().accept(this);
            return derivativeSet;           
        }
        
        public void visitPepaPrefix(PepaPrefix prefix){
            PepaDerivative d = new PepaDerivative(this.current, new PepaAction[]{prefix.getActivity().getAction()},
            new PepaRate[]{prefix.getActivity().getRate()},
            // assuming the model is normed, prefix always contains a PepaVariable
            new PepaVariable[]{(PepaVariable)prefix.getProcess()});            
            derivativeSet.add(d);           
            // reset current
            current = (PepaVariable)prefix.getProcess();            
            if(!this.contains(current)){
                ((PepaProcess)definitions.get(current)).accept(this);
            }
            
        }
        public void visitPepaChoice(PepaChoice choice){
            PepaPrefix prefix;
            List choices = choice.getChoices();
            PepaAction[] actions = new PepaAction[choices.size()];
            PepaRate[] rates = new PepaRate[choices.size()];
            PepaVariable[] targets = new PepaVariable[choices.size()];
            // ok ok, i know the preffered idion here is the iterator
            // but this makes the loop faster and i need to write less code
            for(int i = 0;i<choices.size();i++){
                prefix = (PepaPrefix)choices.get(i);
                actions[i] = prefix.getActivity().getAction();
                rates[i] = prefix.getActivity().getRate();
                targets[i] = (PepaVariable)prefix.getProcess();
            }
            derivativeSet.add(new PepaDerivative(this.current, actions,rates,targets));
            
            // ok derivatives set established now continue finding
            for(int i = 0 ;i<targets.length;i++){
                current = targets[i];
                if(!this.contains(current))
                    ((PepaProcess)definitions.get(current)).accept(this);
            }
        }
        
        /** Used to check if the current definition is already contained within
         *  the current derivatives set */
        private boolean contains(PepaVariable var){
            PepaVariable current;            
            for(Iterator j = derivativeSet.iterator();j.hasNext();){
                current = (PepaVariable)((PepaDerivative)j.next()).getDerivative();                
                if(current.equals(var))
                    return true;
            }
            return false;
        }
        
        
        void visitPepaVariable(PepaVariable var){
            // never need to invoke
        }
        void visitPepaCooperation(PepaCooperation coop){
            // cooperations should not be in definitions
        }
        void visitPepaHide(PepaHide hide){
            // hides should not be in definitions.
        }
    }
}