package eduni.inf.pepa.pepamodel.visitors;

import eduni.inf.pepa.pepamodel.PepaModel;
import eduni.inf.pepa.pepamodel.other.PepaDefinition;
import eduni.inf.pepa.pepamodel.process.*;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;



/** <p>Responsible for norming the pepa model.</p>
 *  The normaliser expects the model to be statically checked.
 *
 *  @author  N.V.Haenel
 *  @since PEPA Workbench 0.9.5
 */
public class ModelNormaliser {
    /** The normed model */
    private PepaModel normed;
    /** Any cooperations in definitions */
    private HashMap cooperations;
    /** Any Hides in definitions */
    private HashMap hide;
    
    /** Creates a new instance of ModelNormaliser */
    public ModelNormaliser() {
    }
    
    /** Normalises a <code>PepaModel</code>
     *
     *  @param model the original <code>PepaModel</code> to be normed.
     *  @return the normed <code>PepaModel</code>
     *
     */
    public PepaModel normaliseModel(PepaModel model){
        normed = new PepaModel();
        cooperations = new HashMap();
        
        DefinitionNormaliser v = new DefinitionNormaliser();
        for(Iterator i = model.getDefinitions().iterator();i.hasNext();){
            v.normalisePepaDefinition((PepaDefinition)i.next());
        }
        
        PepaProcess system = model.getSystemEquation();
	
        
        if(!cooperations.isEmpty())
            //if its just a variable, get the cooperation represented by this
            if(system instanceof PepaVariable){
                system = (PepaProcess)cooperations.get(system);            
            }

            // cooperations to be inserted
            this.expandCooperation((PepaCooperation)system);
        normed.setSystemEquation(system);
        return normed;
    }
    /** Insert any cooperations found in the definition list into
     *  the system equation.
     *
     *  This will stop expanding when the left and right sub-processes
     *  are no longer keys in the cooperations <code>HashMap</code>.
     *
     *  @param coop the cooperation to be expanded
     */
    public void expandCooperation(PepaCooperation coop){
        if(coop.getProcess1() instanceof PepaCooperation)
            expandCooperation((PepaCooperation)coop.getProcess1());
        else if(coop.getProcess1() instanceof PepaVariable){
            if(cooperations.containsKey(coop.getProcess1())){
                coop.setProcess1((PepaProcess)cooperations.get(coop.getProcess1()));
                expandCooperation((PepaCooperation)coop.getProcess1());
            }
        }
        if(coop.getProcess2() instanceof PepaCooperation)
            expandCooperation((PepaCooperation)coop.getProcess1());
        else if(coop.getProcess2() instanceof PepaVariable){
            if(cooperations.containsKey(coop.getProcess2())){
                coop.setProcess2((PepaProcess)cooperations.get(coop.getProcess2()));
                expandCooperation((PepaCooperation)coop.getProcess2());
            }
        }
        
        
    }
    
    /** Re-usable class to normalise a definition within a PEPA model.
     *  Should be re-used to avoid having to create a separate
     *  <code>DefinitionNormaliser</code> for every definition.
     */
    public class DefinitionNormaliser extends AbstractPepaProcessVisitor{
        /** Holds the sequence number of the next Variable to be introduced */
        int count;
        /** the original Definition */
        PepaDefinition definition;
        /** empty constructor to fascilitate instance reuse */
        DefinitionNormaliser(){
        }
        
        /** Normalises a single definition
         *
         *  @param definition the <code>PepaDefinition</code> to be normalised.
         */
        private void normalisePepaDefinition(PepaDefinition definition){
            count = 1;
            this.definition = definition;
            ((PepaProcess)definition.getDefinition()).accept(this);
        }
        
        public void visitPepaPrefix(PepaPrefix prefix){
            if(prefix.getProcess() instanceof PepaVariable){
                if(count==1){
                    //simple definition just pass through
                    normed.addDefinition(definition);
                }
                else{
                    // end of a truncation
                    PepaVariable v = new PepaVariable(definition.getVariable().getIdentifier()+"&"+(count-1));
                    normed.addDefinition(new PepaDefinition(v,prefix));
                }
            }else{
                if(count ==1 ){
                    // start of a truncation
                    PepaVariable  v = new PepaVariable(definition.getVariable().getIdentifier()+"&"+count);
                    // generate new prefix with the identifier
                    PepaPrefix p = new PepaPrefix(prefix.getActivity(),v);
                    // add definition
                    PepaDefinition d = new PepaDefinition(definition.getVariable(),p);
                    normed.addDefinition(d);
                    count++;
                }
                else{
                    // continue a truncation
                    PepaVariable v1 = new PepaVariable(definition.getVariable().getIdentifier()+"&"+(count-1));
                    PepaVariable v2 = new PepaVariable(definition.getVariable().getIdentifier()+"&"+count);
                    // get the activity and construct a new prefix
                    PepaPrefix p = new PepaPrefix(prefix.getActivity(),v2);
                    // add definition
                    normed.addDefinition(new PepaDefinition(v1,p));
                    count++;
                }
                // visit the remainder
                prefix.getProcess().accept(this);
                
            }
        }
        public void visitPepaChoice(PepaChoice choice){      
            PepaVariable choiceIdentifier;
            if(count == 1)  choiceIdentifier = definition.getVariable();                        
            else{
                choiceIdentifier = new PepaVariable(definition.getVariable().getIdentifier()+"&"+(count-1));
                count++;
                
            }
            // make new array List for choices
            ArrayList newChoices = new ArrayList();
            for(Iterator i = choice.getChoices().iterator();i.hasNext();){
                PepaProcess proc = (PepaProcess)i.next();
                PepaProcess newproc = null;
                if(proc instanceof PepaVariable){
                    
                }else if(proc instanceof PepaPrefix){
                    if(((PepaPrefix)proc).getProcess() instanceof PepaVariable){
                        
                    }else{
                        // make a new variable name
                        PepaVariable v = new PepaVariable(definition.getVariable().getIdentifier()+"&"+(count));
                        count++;
                        newproc = new PepaPrefix(((PepaPrefix)proc).getActivity(),v);
                        this.visit(((PepaPrefix)proc).getProcess());
                    }
                }
                // add the old/new process
                newChoices.add(newproc==null?proc:newproc);
            }
            // make a new choice and add to the normed model
            PepaChoice newchoice = new PepaChoice(newChoices);
            normed.addDefinition(new PepaDefinition(choiceIdentifier,newchoice));                
    }
    
    
    void visitPepaVariable(PepaVariable var){
        // should never be called, as we
        // always check for the existence
        // of Variables before truncating any sequential component
        // consider throwing an Excpetion here
    }
    public void visitPepaCooperation(PepaCooperation coop){
        PepaVariable v1;
        PepaVariable v2;
        if (!(coop.getProcess1() instanceof PepaVariable)){
            count++;
            v1 = new PepaVariable(definition.getVariable().getIdentifier()+"&"+(count-1));
            coop.getProcess1().accept(this);
        }else
            v1 = (PepaVariable)coop.getProcess1();
        
        
        if (!(coop.getProcess2() instanceof PepaVariable)){
            count++;
            v2 = new PepaVariable(definition.getVariable().getIdentifier()+"&"+(count-1));
            
            coop.getProcess1().accept(this);
        }else
            v2 = (PepaVariable)coop.getProcess2();
        // add this cooperation to the cooperation HashMap
        // for later insertion into the system equation
        cooperations.put(definition.getVariable(),new PepaCooperation(v1,v2, coop.getCoopSet()));
    }
    void visitPepaHide(PepaHide hide){
        
    }
}
    }
